﻿using System;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Windows.Input;

namespace PackageExplorerViewModel
{
    /// <summary>
    /// A command whose sole purpose is to 
    /// relay its functionality to other
    /// objects by invoking delegates. The
    /// default return value for the CanExecute
    /// method is 'true'.
    /// </summary>
    public class RelayCommand<T> : ICommand
    {
        public RelayCommand(Action<T> execute)
            : this(execute, null)
        {
        }

        /// <summary>
        /// Creates a new command.
        /// </summary>
        /// <param name="execute">The execution logic.</param>
        /// <param name="canExecute">The execution status logic.</param>
        public RelayCommand(Action<T> execute, Predicate<T>? canExecute)
        {
            _execute = execute ?? throw new ArgumentNullException(nameof(execute));
            _canExecute = canExecute;
        }

        [SuppressMessage("Microsoft.Design", "CA1030:UseEventsWhereAppropriate")]
        public void RaiseCanExecuteChanged()
        {
#if WINDOWS
            CommandManager.InvalidateRequerySuggested();
#endif
        }

        [DebuggerStepThrough]
        public bool CanExecute(object? parameter)
        {
            return _canExecute == null || _canExecute((T)parameter!);
        }

        public event EventHandler? CanExecuteChanged
        {
            add
            {
#if WINDOWS
                if (_canExecute != null)
                {
                    CommandManager.RequerySuggested += value;
                }
#endif
            }
            remove
            {
#if WINDOWS
                if (_canExecute != null)
                {
                    CommandManager.RequerySuggested -= value;
                }
#endif
            }
        }

        public void Execute(object? parameter)
        {
            _execute((T)parameter!);
        }

        private readonly Predicate<T>? _canExecute;
        private readonly Action<T> _execute;
    }

    /// <summary>
    /// A command whose sole purpose is to 
    /// relay its functionality to other
    /// objects by invoking delegates. The
    /// default return value for the CanExecute
    /// method is 'true'.
    /// </summary>
    public class RelayCommand : ICommand
    {
        #region Constructors

        /// <summary>
        /// Creates a new command that can always execute.
        /// </summary>
        /// <param name="execute">The execution logic.</param>
        public RelayCommand(Action execute)
            : this(execute, null)
        {
        }

        /// <summary>
        /// Creates a new command.
        /// </summary>
        /// <param name="execute">The execution logic.</param>
        /// <param name="canExecute">The execution status logic.</param>
        public RelayCommand(Action execute, Func<bool>? canExecute)
        {
            _execute = execute ?? throw new ArgumentNullException(nameof(execute));
            _canExecute = canExecute;
        }

        #endregion // Constructors

        #region ICommand Members

        [DebuggerStepThrough]
        public bool CanExecute(object? parameter)
        {
            return _canExecute == null || _canExecute();
        }

        public event EventHandler? CanExecuteChanged
        {
            add
            {
#if WINDOWS
                if (_canExecute != null)
                {
                    CommandManager.RequerySuggested += value;
                }
#endif
            }
            remove
            {
#if WINDOWS
                if (_canExecute != null)
                {
                    CommandManager.RequerySuggested -= value;
                }
#endif
            }
        }

        public void Execute(object? parameter)
        {
            _execute();
        }

        #endregion

        #region Fields

        private readonly Func<bool>? _canExecute;
        private readonly Action _execute;

        #endregion // Fields

        [SuppressMessage("Microsoft.Design", "CA1030:UseEventsWhereAppropriate"),
         SuppressMessage("Microsoft.Performance", "CA1822:MarkMembersAsStatic")]
        public void RaiseCanExecuteChanged()
        {
#if WINDOWS
            CommandManager.InvalidateRequerySuggested();
#endif
        }
    }
}
